import logging

from django.conf import settings
from django.core.mail.backends.base import BaseEmailBackend
from django.utils.module_loading import import_string

from django_darthmail import DarthMailError
from django_darthmail.message import EmailMultiAlternatives
from django_darthmail.client import DarthMailClient

logger = logging.getLogger(__name__)


class DarthMailBackend(BaseEmailBackend):
    def __init__(self, url=None, token=None, fallback=None, *args, **kwargs):
        self.client = DarthMailClient(
            url or settings.EMAIL_DARTHMAIL_URL,
            token or settings.EMAIL_DARTHMAIL_TOKEN,
        )
        try:
            self.fallback = import_string(fallback or settings.EMAIL_DARTHMAIL_FALLBACK)(*args, **kwargs)
        except ImportError:
            self.fallback = None
        super(DarthMailBackend, self).__init__(*args, **kwargs)

    def send_messages(self, email_messages):
        for msg in email_messages:
            # EmailMessage might try to cache the connection; not useful in our case.
            # Also breaks fallback logic.
            msg.connection = None

            if isinstance(msg, EmailMultiAlternatives) and not set(msg.metadata.keys()).intersection(
               settings.EMAIL_DARTHMAIL_REQUIRED_METADATA_FIELDS):
                # we try to avoid sending emails that would be "invisible"
                raise DarthMailError(
                    'must include one metadata field of %s' % settings.EMAIL_DARTHMAIL_REQUIRED_METADATA_FIELDS
                )

            if isinstance(msg, EmailMultiAlternatives) and (not settings.DEBUG or self.client.ping()):
                logger.info('submitting %s to darthmail', msg)
                try:
                    self.client.submit(msg)
                except Exception:
                    if not self.fail_silently:
                        raise
            elif self.fallback:
                logger.info('submitting %s through fallback backend (%s)', msg, self.fallback)
                self.fallback.send_messages(email_messages)
            else:
                raise DarthMailError('no fallback configured for non-darthmail messages. Set EMAIL_DARTHMAIL_FALLBACK.')



from django.utils import timezone

from herocentral.core.management.base import DevCommand
from django_darthmail.message import EmailMultiAlternatives


class Command(DevCommand):
    help = 'test darthmail'

    def handle(self, *args, **options):
        em = EmailMultiAlternatives(
            subject='darthmail test via HC %s' % timezone.now(),
            body='test email via HC %s' % timezone.now(),
            from_email='HeroCentral@regiohelden.de',
            to=['foo@regiohelden.de', 'bar@regiohelden.de'],
            cc=[],
            bcc=['baz@regiohelden.de'],
            reply_to=['noreply@regiohelden.de'],
            headers={'X-Test': 'true'},
            attachments=[],
            metadata={
                'test': 'true',
                'id': 42,
            },
        )
        em.send()



from email.mime.base import MIMEBase

from django.core.serializers.json import DjangoJSONEncoder

from django_darthmail.message import EmailMultiAlternatives


def _b64encode(content):
    from base64 import b64encode
    if isinstance(content, str):
        content = content.encode('utf8')
    return b64encode(content).decode('ascii')


class EmailSerializer(DjangoJSONEncoder):
    # some parameter names and the property names don't match in django.core.mail.message.EmailMessage;
    # we have to fix the mismatch when serializing.
    RENAME_FIELDS = {
        'extra_headers': 'headers',
    }

    def default(self, obj):
        res = dict()
        if isinstance(obj, EmailMultiAlternatives):
            res['alternatives'] = list(map(self._encodeTupleOrMime, obj.alternatives))
            res['attachments'] = list(map(self._encodeTupleOrMime, obj.attachments))
            keys = ['body', 'bcc', 'cc', 'extra_headers', 'from_email', 'reply_to', 'subject', 'to', 'metadata',
                    'send_at', 'do_not_send', 'force_send', 'unique_hash']
            for key in keys:
                dst_key = self.RENAME_FIELDS.get(key, key)
                try:
                    res[dst_key] = self._normalize(key, obj.__getattribute__(key))
                except AttributeError:
                    pass  # leave non-existing attributes empty
            return res
        elif isinstance(obj, MIMEBase):
            if obj.is_multipart():
                raise NotImplementedError('cannot serialize multipart MIME yet')

            headers = dict()
            for key in obj.keys():
                # these are managed by the MIME* type itself
                if key not in ['Content-Type', 'MIME-Version', 'Content-Transfer-Encoding']:
                    headers[key] = obj[key]

            # is this even a thing? I'm pretty sure we don't use anything else
            if obj['Content-Transfer-Encoding'] != 'base64':
                raise NotImplementedError(
                    'MIME object with unsupported transfer-encoding: %s' % obj['Content-Transfer-Encoding']
                )

            return dict(
                filename=obj.get_filename(),
                content=obj.get_payload(),  # already base64 encoded; see above
                mimetype=obj.get_content_type(),
                headers=headers,
            )
        return super(EmailSerializer, self).default(obj)

    def _normalize(self, key, val):
        if key == 'metadata':
            # The API expects metadata to be a dict with lists as values. To make the client simpler, we wrap scalars
            # in a list.
            for k, v in val.items():
                if not isinstance(v, list) and not isinstance(v, tuple):
                    val[k] = [v]
        elif key == 'subject':
            # cut subject to 200 characters as Darthmail enforces this limit
            val = val[:200]
        return val

    def _encodeTupleOrMime(self, attachment):
        if isinstance(attachment, tuple):
            if len(attachment) == 2:
                return dict(
                    content=_b64encode(attachment[0]),
                    mimetype=attachment[1],
                )
            elif len(attachment) == 3:
                return dict(
                    filename=attachment[0],
                    content=_b64encode(attachment[1]),
                    mimetype=attachment[2],
                )
            else:
                raise TypeError('unsupported attachment/alternative format; tuple with %d elements' % len(attachment))
        elif isinstance(attachment, MIMEBase):
            return self.default(attachment)
        else:
            raise TypeError('unsupported attachment type %s' % type(attachment))



__version__ = "0.4.0"


class DarthMailError(Exception):
    def __init__(self, inner, *args):
        self.inner = inner
        super(DarthMailError, self).__init__(inner, *args)

    def __str__(self):
        try:
            return "%s: %s" % (super(DarthMailError, self).__str__(), self.inner.response.content)
        except AttributeError:
            return super(DarthMailError, self).__str__()


class DarthMailDuplicateError(DarthMailError):
    pass



import json
import logging
from urllib.parse import urljoin

from django.conf import settings
import requests

from django_darthmail import DarthMailError, DarthMailDuplicateError
from django_darthmail.message import EmailMultiAlternatives
from django_darthmail.serializers import EmailSerializer

API_BASE_URL = 'api/v1/'

logger = logging.getLogger(__name__)


class DarthMailClient(object):
    '''
    This client should only be used via the DarthMailBackend
    '''

    def __init__(self, url=None, token=None, endpoint='emailmsg/', headers={}):
        self.api_url = url or settings.EMAIL_DARTHMAIL_URL
        self.token = token or settings.EMAIL_DARTHMAIL_TOKEN
        # normalize to ensure last slash
        self.endpoint_url = urljoin(API_BASE_URL, endpoint).strip('/') + '/'
        self.headers = headers

    def _get_headers(self, extra=dict()):
        res = {
            'Authorization': 'Token %s' % self.token,
        }
        res.update(self.headers)
        res.update(extra)
        return res

    def _post(self, obj, url=None):
        url = urljoin(urljoin(self.api_url, self.endpoint_url), url)
        if isinstance(obj, EmailMultiAlternatives):
            return requests.post(
                url,
                headers=self._get_headers(extra={'Content-Type': 'application/json'}),
                data=json.dumps(obj, cls=EmailSerializer),
            )
        return requests.post(url, headers=self._get_headers(), json=obj)

    def _get(self, url=None, **kwargs):
        url = urljoin(urljoin(self.api_url, self.endpoint_url), url)
        for k, v in kwargs.items():
            if isinstance(v, dict):
                kwargs[k] = ['%s:%s' % (k2, v2) for k2, v2 in v.items()]
        return requests.get(url, headers=self._get_headers(), params=kwargs)

    def _delete(self, url=None):
        url = urljoin(urljoin(self.api_url, self.endpoint_url), url)
        return requests.delete(url, headers=self._get_headers())

    def ping(self):
        url = urljoin(self.api_url, self.endpoint_url)
        try:
            requests.options(url, headers=self._get_headers()).raise_for_status()
            return True
        except Exception:
            return False

    def submit(self, msg):
        '''
        Submit message to darthmail
        '''

        res = _process_response(self._post(msg))
        msg.darthmail_id = res['id']  # TODO: used for syncing, can be removed when migration is done (see #104995)
        return res

    def resend(self, pk, to=None, cc=None, bcc=None, reply_to=None, from_email=None):
        '''
        Resend a given email message (identified by PK), optionally overwriting its TO, CC, BCC, REPLY TO and FROM
        fields.
        '''
        url = '%s/resend/' % pk
        overrides = dict()
        if to:
            overrides['to'] = to
        if cc:
            overrides['cc'] = cc
        if bcc:
            overrides['bcc'] = bcc
        if reply_to:
            overrides['reply_to'] = reply_to
        if from_email:
            overrides['from_email'] = from_email
        return _process_response(self._post(overrides, url=url))

    def filter(self, to=[], cc=[], bcc=[], from_email=None, metadata=dict(), any_recipient=[], latest_status=[],
               send_at_before=None, send_at_after=None, unique_hash=None):
        '''
        Search DarthMail for messages matching the provided criteria.

        All parameters will be joined by AND, including the individual items of parameters that accept lists or dicts.
        '''
        return EmailIterator(
            self,
            to=self._format_list(to),
            cc=self._format_list(cc),
            bcc=self._format_list(bcc),
            from_email=from_email,
            metadata=self._format_list_dict(metadata),
            any_recipient=self._format_list(any_recipient),
            latest_status=latest_status,
            send_at_before=send_at_before,
            send_at_after=send_at_after,
            unique_hash=unique_hash,
        )

    def _format_list(self, lst):
        return ','.join(lst)

    def _format_list_dict(self, dct):
        lst = list()
        for k, vl in dct.items():
            if not isinstance(vl, list) and not isinstance(vl, tuple):
                vl = [vl]
            for v in vl:
                lst.append('%s:%s' % (k, v))

        return ','.join(lst)


class EmailIterator(object):
    def __init__(self, client, to=[], cc=[], bcc=[], from_email=None, metadata=dict(), any_recipient=[],
                 latest_status=[], send_at_before=None, send_at_after=None, unique_hash=None):
        self.client = client
        self.response = _process_response(
            client._get(to=to, cc=cc, bcc=bcc, from_email=from_email, metadata=metadata, any_recipient=any_recipient,
                        latest_status=latest_status, send_at_before=send_at_before, send_at_after=send_at_after,
                        unique_hash=unique_hash)
        )

    def __len__(self):
        return self.response['count']

    def __iter__(self):
        while True:
            for msg in self.response['results']:
                yield msg
            if self.response['next']:
                self.response = _process_response(
                    requests.get(self.response['next'], headers=self.client._get_headers())
                )
                continue
            else:
                return


def _process_response(response):
    try:
        content = response.json()
    except requests.exceptions.JSONDecodeError:
        content = response.content

    try:
        response.raise_for_status()
    except requests.HTTPError:
        if response.status_code == 409 and content.get('code', None) == 'duplicate':
            raise DarthMailDuplicateError(content)
        raise DarthMailError(content)
    return content



import mimetypes
from email.mime.image import MIMEImage
from pathlib import Path
from typing import IO

from django.core.mail import EmailMultiAlternatives as DjangoEmailMultiAlternatives
from PIL import Image


class EmailMultiAlternatives(DjangoEmailMultiAlternatives):
    def __init__(self, *args, **kwargs):
        """
        Extension of django's EmailMultiAlternatives supporting the following extra parameters: metadata, send_at,
        unique_hash and force_send

        @param metadata: a dict where each value is either a scalar (int, str, etc) or a list of scalars.
        @param send_at: date when to actually send the email.
        @param force_send: boolean indicating if email should be send even if it was already sent before
        @param unique_hash: string containing unique hash for duplicates detection
        """
        self.metadata = kwargs.pop('metadata', dict())
        self.send_at = kwargs.pop('send_at', None)
        self.force_send = kwargs.pop('force_send', False)
        self.unique_hash = kwargs.pop('unique_hash', None)

        super(EmailMultiAlternatives, self).__init__(*args, **kwargs)

    def attach_inline_image(self, filepath: str, cid: str) -> None:
        if not Path(filepath).exists():
            # TODO: shouldn't we raise something?
            return
        with Path(filepath).open(mode='rb') as fp:
            image_data = fp.read()
            image = Image.open(fp)
            _, mime_format = image.get_format_mimetype().split('/')
            msg_img = MIMEImage(image_data, _subtype=mime_format)
            msg_img.add_header('Content-ID', '<{}>'.format(cid))
            self.attach(msg_img)

    def add_image(self, directory: str, filename: str, alias: str) -> None:
        '''
        Wrapper for backward-compat with notification.models.EmailMultiAlternativesWithImages
        '''
        filepath = Path(directory, filename)
        self.attach_inline_image(filepath.as_posix(), alias)

    def add_attachment(self, _file: IO) -> None:
        content_type, encoding = mimetypes.guess_type(_file.name)
        self.attach(Path(_file.name).name, _file.read(), content_type)



import os
import re
from os.path import join, dirname
from setuptools import setup, find_namespace_packages


def get_version(*file_paths):
    """Retrieves the version from the given path"""
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    with open(filename) as f:
        version_file = f.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def long_description():
    """Return long description from README.md if it's present
    because it doesn't get installed."""
    try:
        with open(join(dirname(__file__), "README.md")) as f:
            return f.read()
    except IOError:
        return ""


setup(
    name='django-darthmail',
    packages=[
        'django_darthmail',
    ],
    version=get_version("django_darthmail", "__init__.py"),
    description="""Client for the DarthMail project""",
    long_description=long_description(),
    long_description_content_type='text/markdown',
    author='RegioHelden GmbH',
    author_email='entwicklung@regiohelden.de',
    url='https://github.com/regiohelden/django-darthmail',

    include_package_data=True,
    install_requires=[
        "django>=4.2",
        "requests>=2.31.0",
    ],
    license="BSD-3-Clause",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Communications :: Email',
    ],
)
